//		Copyright © 2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		MIDI consumer class for the Csound Server.

#include "CsSvMidiConsumer.h"

#include <cstring>

#include <Autolock.h>
#include <List.h>
#include <Locker.h>
#include <MidiConsumer.h>
#include <OS.h>

#include "CsSvInterface.h"

//#define DEBUGP
#include "dprintf.h"


namespace
{
	const char kWatcherThreadName[] = "Sword of Damocles";
}

BList CsoundServerMidiConsumer::smConsumerList;
BLocker CsoundServerMidiConsumer::smConsumerListLocker("consumer list");

CsoundServerMidiConsumer::CsoundServerMidiConsumer(
	const char *pName, const ServerMidiPort *pPort)
:	inherited(pName),
	mPort(pPort->mPort),
	mClientThread(pPort->mClientThread)
{
	// Start a thread which monitors the client.
	thread_id watcherThread =
		spawn_thread(WatcherThreadStub, kWatcherThreadName, B_NORMAL_PRIORITY, this);
	if (watcherThread >= B_OK) {
		// Start it running.
		if (resume_thread(watcherThread) < B_OK) {
			// Dang.  What to do now?
			kill_thread(watcherThread);
		}
	}

	// Register the instance locally.
	BAutolock lock(smConsumerListLocker);
	smConsumerList.AddItem(this);
}

CsoundServerMidiConsumer::~CsoundServerMidiConsumer(void)
{
	// empty
	DPRINTF(("Consumer destructor called\n");)
}

void
CsoundServerMidiConsumer::Data(uchar *pData, size_t length, bool, bigtime_t time)
{
	// Pass data to the client.
	DPRINTFZ(("sending MIDI to Csound %d bytes at %.3f sec\n",
	 length, time/1000000.0);)
	(void)write_port(mPort, 0, pData, length);
}

// Find an active consumer (needed to work around a bug in BMidiRoster).
CsoundServerMidiConsumer *
CsoundServerMidiConsumer::Find(const char *pName)
{
	BAutolock lock(smConsumerListLocker);

	const int32 nItems = smConsumerList.CountItems();
	for (int32 i = 0; i < nItems; ++i) {
		CsoundServerMidiConsumer *pConsumer =
			static_cast<CsoundServerMidiConsumer *>(smConsumerList.ItemAt(i));
		if (pConsumer && !strcmp(pName, pConsumer->Name())) {
			pConsumer->Acquire();
			return pConsumer;
		}
	}
	
	return 0;
}

// Thread functions.
int32
CsoundServerMidiConsumer::WatcherThreadStub(void *pData)
{
	return ((CsoundServerMidiConsumer *)pData)->WatcherThread();
}

int32
CsoundServerMidiConsumer::WatcherThread(void)
{
	// Wait for the client to terminate.
	status_t dummy;
	wait_for_thread(mClientThread, &dummy);
	
	// Unregister and release the MIDI consumer.
	//Unregister();	-- I believe this is buggy in Maui...
	DPRINTF(("Releasing consumer\n");)
	{
		// Unregister the instance locally.
		BAutolock lock(smConsumerListLocker);
		smConsumerList.RemoveItem(this);
	}
	Release();

	// Done.	
	return 0;
}
