//		Copyright © 1997–2001  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Application class for the Csound Server.

#include "CsSvApplication.h"

#include <cstdio>
#include <cstring>

#include <Alert.h>
#include <AppFileInfo.h>
#include <Application.h>
#include <File.h>
#include <Bitmap.h>
#include <Message.h>
#include <MidiEndpoint.h>
#include <MidiProducer.h>
class BMidiConsumer;
#include <MidiRoster.h>
#include <InterfaceDefs.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <Roster.h>
#include <String.h>

#include "CsSvClient.h"
#include "CsSvInterface.h"
#include "CsSvMidiConsumer.h"
#include "CsSvMidiProducer.h"
#include "CsSvMidiSynthConsumer.h"
#include "CsSvSoundPlayer.h"
#include "CsSvStringReader.h"

// Constant data.
namespace
{
	const int32 kPortCapacity = 100;
	const char kServerThreadName[] = "Csound server thread";
	const char kAudioThreadName[] = "Csound audio thread";
	const char kAudioSemName[] = "Sound player finished";
	const char kLargeIconName[] = "be:large_icon";
	const char kMiniIconName[] = "be:mini_icon";
	const char kSynthConsumerName[] = "Csound:BMidiSynth";
	const char kMidiEndpointNamePrefix[] = "Csound:";
	enum {
		kLargeIconType = 'ICON',
		kMiniIconType = 'MICN'
	};
}

// BApplication requirements.
CsoundServerApplication::CsoundServerApplication(void)
:	inherited(kServerSignature),
	mServerPort(B_ERROR),
	mServerThread(B_ERROR),
	mServerFinished(B_ERROR),
	mLargeIcon(BRect(0, 0, 31, 31), B_CMAP8),
	mMiniIcon(BRect(0, 0, 15, 15), B_CMAP8),
	mpSynth(new CsoundServerMidiSynthConsumer(kSynthConsumerName))
{
	// Get the application's icons.
	app_info appInfo;
	if (GetAppInfo(&appInfo) == B_OK) {
		BFile file(&(appInfo.ref), B_READ_ONLY);
		if (file.InitCheck() == B_OK) {
			BAppFileInfo appFileInfo(&file);
			if (appFileInfo.InitCheck() == B_OK) {
				appFileInfo.GetIcon(&mLargeIcon, B_LARGE_ICON);
				appFileInfo.GetIcon(&mMiniIcon, B_MINI_ICON);
			}
		}
	}
	
	// Register the software synthesizer.
	AddIcons(*mpSynth);
	mpSynth->Register();
}

CsoundServerApplication::~CsoundServerApplication(void)
{
	// Get rid of the software synthesizer.
	mpSynth->Unregister();
	mpSynth->Release();
	mpSynth = 0;
	
	// Stop the server thread and clean up.
	if (mServerPort >= B_OK) {
		delete_port(mServerPort);
		mServerPort = B_ERROR;
	}
	if (mServerThread >= B_OK) {
		kill_thread(mServerThread);
		mServerThread = B_ERROR;
	}
	if (mServerFinished >= B_OK) {
		delete_sem(mServerFinished);
		mServerFinished = B_ERROR;
	}
}

// Hook functions.
void
CsoundServerApplication::AboutRequested(void)
{
	(new BAlert("About",
					"The Csound Server\n"
					"pleases you, honoured user\n"
					"with sounds and graphics.",
					"OK"))->Go();
}

bool
CsoundServerApplication::QuitRequested(void)
{
	if (mServerPort >= B_OK) {
		// Tell the server thread to quit, then wait until it has done so.
		if (write_port(mServerPort, kServerQuit, 0, 0) >= B_OK) {
			acquire_sem(mServerFinished);
			mServerThread = B_ERROR;
		}
	}
	return true;
}

void
CsoundServerApplication::ReadyToRun(void)
{
	// Create the communications port.
	mServerPort = create_port(kPortCapacity, kServerPortName);
	if (mServerPort >= B_OK) {
		// Create the semaphore.
		mServerFinished = create_sem(0, "End of server thread");
		if (mServerFinished >= B_OK) {
			// Spawn the server thread.
			mServerThread =
				spawn_thread(ServerThreadStub, kServerThreadName, B_NORMAL_PRIORITY, this);
			if (mServerThread >= B_OK) {
				// Start it running.
				if (resume_thread(mServerThread) < B_OK) {
					// Dang.
					kill_thread(mServerThread);
					mServerThread = B_ERROR;
				}
			}
			
			if (mServerThread < B_OK) {
				// Drat.
				delete_sem(mServerFinished);
				mServerFinished = B_ERROR;
			}
		}
		
		if (mServerFinished < B_OK) {
			// Blast.
			delete_port(mServerPort);
			mServerPort = B_ERROR;
		}
	}
	
	if (mServerPort < B_OK) {
		// Buggrit, buggrit, millennium hand and shrimp!
		(new BAlert("", "Cannot start Csound Server.", "Duh" B_UTF8_ELLIPSIS))
			->Go();
		PostMessage(B_QUIT_REQUESTED);
	}
}

// Handle the various client requests.
int32
CsoundServerApplication::ServerThreadStub(void *pData)
{
	return ((CsoundServerApplication *)pData)->ServerThread();
}

int32
CsoundServerApplication::ServerThread(void) const
{
	// Read messages from the port and act on them.
	bool finished = false;
	while (!finished) {
		int32 msgCode;
		char msgBuffer[ServerBufferSize];
		if (read_port(mServerPort, &msgCode, msgBuffer, sizeof(msgBuffer))
			 >= B_OK) {
			switch (msgCode) {
			
			case kServerNewAudioInPort:
				// Accept a new audio connection for input.
				// ### Currently, audio input is not supported.  Close the
				// ### connection immediately to let the client know.
				delete_port(((ServerAudioPort *)msgBuffer)->mPort);
				break;
			
			case kServerNewAudioOutPort:
				// Accept a new audio connection for output.
				NewAudioOutThread((ServerAudioPort *)msgBuffer);
				break;
			
			case kServerNewClientPort:
				// Accept a new window connection.
				NewClient((ServerClientPort *)msgBuffer);
				break;
			
			case kServerNewMidiInPort:
				// Accept a new MIDI connection for input.
				NewMidiInput((ServerMidiPort *)msgBuffer);
				break;
			
			case kServerNewMidiOutPort:
				// Accept a new MIDI connection for output.
				NewMidiOutput((ServerMidiPort *)msgBuffer);
				break;
			
			case kServerQuit:
				// ### MUST CLEAN UP
				finished = true;
				break;
			
			default:
				// Sounds foreign to me.
				sprintf(msgBuffer, "Csound Server: 0x%08lx Huh?", msgCode);
				(new BAlert("", msgBuffer, "Duh" B_UTF8_ELLIPSIS))->Go();
				break;
			}
		} else {
			// Cannot read from port.
			(new BAlert("", "Csound Server: Arrgh!", "Oops" B_UTF8_ELLIPSIS))->Go();
			finished = true;
		}
	}
	
	// Signal that we're finished.
	release_sem(mServerFinished);
	return 0;
}

void
CsoundServerApplication::NewAudioOutThread(ServerAudioPort *pDesc) const
{
	// Copy the descriptor (there is a race condition).
	ServerAudioPort *pNewDesc = new ServerAudioPort;
	*pNewDesc = *pDesc;
	
	// Spawn the audio thread and start it running.
	thread_id audioThread =
		spawn_thread(AudioOutThread, kAudioThreadName, B_NORMAL_PRIORITY, pNewDesc);
	if (audioThread >= B_OK) {
		if (resume_thread(audioThread) < B_OK) {
			kill_thread(audioThread);
		}
	}
}

int32
CsoundServerApplication::AudioOutThread(void *pData)
{
	// ### This is preliminary, using a port to transfer audio data
	// ### and a BSoundPlayer to play it.  It should be rewritten to
	// ### use a more efficient means of transfer (shared memory?),
	// ### and media nodes for playing and recording.  That's a lot of work.

	ServerAudioPort *pDesc = (ServerAudioPort *)pData;

	// All the work is done by the sound player class while this thread
	// blocks on a semaphore.  When the client closes the audio port,
	// the sound player releases the semaphore and we stop and delete it.
	sem_id playerFinished = create_sem(0, kAudioSemName);
	if (playerFinished < B_OK) {
		// Close port to let the client know that something went wrong.
		delete_port(pDesc->mPort);

		return playerFinished;
	}

	CsoundServerSoundPlayer *pPlayer =
		new CsoundServerSoundPlayer(*pDesc, playerFinished);
	status_t status = pPlayer->InitCheck();
	if (status == B_OK) {
		status = pPlayer->Start();
	}
	if (status == B_OK) {
		// Wait until the client has finished, then stop playing.
		acquire_sem(playerFinished);
		pPlayer->Stop();
	} else {
		// Close port to let the client know that something went wrong.
		delete_port(pDesc->mPort);
	}
	
	// Clean up.
	delete pPlayer;
	delete_sem(playerFinished);
	delete pDesc;
	
	return status;
}

void
CsoundServerApplication::NewClient(ServerClientPort *pDesc) const
{
	// Create a new client monitor and start it running.
	(new CsoundServerClient(pDesc))->Run();
}

void
CsoundServerApplication::NewMidiInput(ServerMidiPort *pDesc) const
{
	// Read the name of the connection from the client.
	CsoundServerStringReader nameReader(pDesc->mPort);
	BString consumerName = kMidiEndpointNamePrefix;
	consumerName += nameReader;
		
	// Create a new MIDI consumer for handling this connection.
	CsoundServerMidiConsumer *pConsumer =
		new CsoundServerMidiConsumer(consumerName.String(), pDesc);
	AddIcons(*pConsumer);
	pConsumer->Register();

	// If there is a known MIDI producer with the given name, connect it
	// to our consumer.
	if (BMidiProducer *pProducer = FindProducer(nameReader)) {
		pProducer->Connect(pConsumer);
		pProducer->Release();
	}
}

void
CsoundServerApplication::NewMidiOutput(ServerMidiPort *pDesc) const
{
	// Read the name of the connection from the client.
	CsoundServerStringReader nameReader(pDesc->mPort);
	BString producerName = kMidiEndpointNamePrefix;
	producerName += nameReader;		
	
	// Create a new MIDI producer for handling this connection.
	CsoundServerMidiProducer *pProducer =
		new CsoundServerMidiProducer(producerName.String(), pDesc);
	AddIcons(*pProducer);
	pProducer->Register();
	
	// Start processing data.
	if (pProducer->Run() < B_OK) {
		pProducer->Unregister();
		pProducer->Release();
		
		(new BAlert("",
						"Csound Server: Problems opening MIDI output.",
						"Sorry" B_UTF8_ELLIPSIS)
		)->Go();
		return;
	}

	// If there is a known MIDI consumer with the given name, connect
	// our producer to it.
	if (BMidiConsumer *pConsumer = FindConsumer(nameReader)) {
		pProducer->Connect(pConsumer);
		pConsumer->Release();
	}
}

// Add icons to a MIDI endpoint.
void
CsoundServerApplication::AddIcons(BMidiEndpoint &endpoint) const
{
	// The icons are stored as a property, by convention.
	BMessage msg;
	if (endpoint.GetProperties(&msg) == B_OK) {

		// Add icon properties.
		if (!msg.HasData(kLargeIconName, kLargeIconType)) {
			msg.AddData(kLargeIconName, kLargeIconType,
							mLargeIcon.Bits(), mLargeIcon.BitsLength());
		} else {
			msg.ReplaceData(kLargeIconName, kLargeIconType,
								 mLargeIcon.Bits(), mLargeIcon.BitsLength());
		}
		if (!msg.HasData(kMiniIconName, kMiniIconType)) {
			msg.AddData(kMiniIconName, kMiniIconType,
							mMiniIcon.Bits(), mMiniIcon.BitsLength());
		} else {
			msg.ReplaceData(kMiniIconName, kMiniIconType,
								 mMiniIcon.Bits(), mMiniIcon.BitsLength());
		}
		
		endpoint.SetProperties(&msg);
	} 
}

// Find MIDI endpoints by name.
BMidiConsumer *
CsoundServerApplication::FindConsumer(const char *pName) const
{
	if (BMidiRoster *pRoster = BMidiRoster::MidiRoster()) {
		// Iterate through all consumers, stopping when a name matches ours.
		int32 id = 0;
		while (BMidiConsumer *pConsumer = pRoster->NextConsumer(&id)) {
			if (!strcmp(pName, pConsumer->Name())) {
				return pConsumer;
			}
			pConsumer->Release();
		}
	}
	
	// Work around a bug in BMidiRoster.
	if (BMidiConsumer *pConsumer = CsoundServerMidiConsumer::Find(pName)) {
		return pConsumer;
	}
	if (!strcmp(pName, kSynthConsumerName)) {
		mpSynth->Acquire();
		return mpSynth;
	}
	
	return 0;
}

BMidiProducer *
CsoundServerApplication::FindProducer(const char *pName) const
{
	if (BMidiRoster *pRoster = BMidiRoster::MidiRoster()) {
		// Iterate through all consumers, stopping when a name matches ours.
		int32 id = 0;
		while (BMidiProducer *pProducer = pRoster->NextProducer(&id)) {
			if (!strcmp(pName, pProducer->Name())) {
				return pProducer;
			}
			pProducer->Release();
		}
	}

	// Work around a bug in BMidiRoster.
	if (BMidiProducer *pProducer = CsoundServerMidiProducer::Find(pName)) {
		return pProducer;
	}

	return 0;
}
