//		Copyright © 2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Sound player class for the Csound Server.

#include "CsSvSoundPlayer.h"

#include <MediaDefs.h>
#include <OS.h>
#include <SoundPlayer.h>

#include "soundio.h"

#include "AudioConversions.h"
#include "CsSvInterface.h"

//#define DEBUGP
#include "dprintf.h"

CsoundServerSoundPlayer::CsoundServerSoundPlayer(
	const ServerAudioPort &port, sem_id finished)
:	inherited(ConstructFormat(port, mFormat), "Csound", PlayStub, 0, this),
	mPort(port.mPort),
	mSampleFormat(port.mSampleFormat),
	mBytesPerSample(BytesPerSample(port.mSampleFormat)),
	mBufferSize(port.mBufferSize),
	mpBuffer(new char[port.mBufferSize]),
	mBufferStart(0),
	mBufferLimit(0),
	mFinished(finished)
{
	// empty
}

CsoundServerSoundPlayer::~CsoundServerSoundPlayer(void)
{
	delete [] mpBuffer;
	mpBuffer = 0;
}

media_raw_audio_format *
CsoundServerSoundPlayer::ConstructFormat(const ServerAudioPort &port,
													  media_raw_audio_format &format)
{
	format.frame_rate = port.mSampleRate;
	format.channel_count = port.mNumChannels;
	format.format = media_raw_audio_format::B_AUDIO_FLOAT;
	format.byte_order = B_MEDIA_HOST_ENDIAN;
	format.buffer_size = port.mBufferSize
							 * sizeof(float)
							 / BytesPerSample(port.mSampleFormat);
	return &format;
}

// How many bytes CSound outputs per sample.
size_t
CsoundServerSoundPlayer::BytesPerSample(int format)
{
	switch (format) {
	
	case AE_CHAR:
	case AE_UNCH:
	case AE_ULAW:
	case AE_ALAW:
		return 1;
		
	case AE_SHORT:
		return 2;

	default:
		return 4;
	}
}

void
CsoundServerSoundPlayer::PlayStub(void *pCookie, void *pBuffer, size_t size,
											 const media_raw_audio_format &format)
{
	(static_cast<CsoundServerSoundPlayer *>(pCookie))
		->Play(pBuffer, size, format);
}

void
CsoundServerSoundPlayer::Play(void *pBuffer, size_t size,
										const media_raw_audio_format &format)
{
	if (format.format != media_raw_audio_format::B_AUDIO_FLOAT) {
		// This can't happen.
		return;
	}

	if (!mpBuffer) {
		// Work around an apparent BeOS bug (Play() called after deletion).
		return;
	}

	size_t toRead = size / sizeof(float);
	float *pSamples = static_cast<float *>(pBuffer);
	#ifdef DEBUGP
		static int64 lastchunk;
		bool show_read=true;
		if (lastchunk >=0)
			printf("Playing to buffer size %d -- %d samples at %Ld, last arr %Ld\n",
			 size, toRead, system_time(), lastchunk);
	#endif

	while (toRead > 0) {
		size_t bufferSize = mBufferLimit - mBufferStart;
		DPRINTFZ(("   converting %d bytes\n", bufferSize);)
		if (bufferSize > toRead * mBytesPerSample) {
			ConvertSamples(mpBuffer + mBufferStart, pSamples, toRead);
			mBufferStart += toRead * mBytesPerSample;
			toRead -= toRead;

		} else {
			size_t nSamples = bufferSize / mBytesPerSample;
			ConvertSamples(mpBuffer + mBufferStart, pSamples, nSamples);
			pSamples += nSamples;
			toRead -= nSamples;

			int32 dummy;
			const ssize_t nRead = read_port(mPort, &dummy, mpBuffer, mBufferSize);
			#ifdef DEBUGP
			int64 thistime = system_time();
			if (show_read && nRead>=0) {
				printf("read %d bytes from port at %Ld\n", nRead, thistime);
				show_read = false;
			}
			lastchunk = nRead>=0? thistime : -1;
			#endif
			if (nRead < 0) {
				// Port closed - stop playing.
				while (toRead-- > 0) {
					*pSamples++ = 0.0f;
				}
				mBufferLimit = 0;
				release_sem(mFinished);
				return;
			}
			mBufferStart = 0;
			mBufferLimit = (size_t)nRead;
		}
	}
	#ifdef DEBUGP
		if (lastchunk >=0)
			printf("Dispatching buffer at %Ld\n", system_time());
	#endif
}

void
CsoundServerSoundPlayer::ConvertSamples(
	const char *pBuffer, float *pOut, size_t nSamples)
{
	switch (mSampleFormat) {
	
	case AE_CHAR:
		DPRINTFZ(("Converting %d samples -- AE_CHAR\n", nSamples);)
		{	const char *pSample = pBuffer;
			const float reciprocal = 1.0f / 128.0f;
			while (nSamples--) {
				*pOut++ = (float)*pSample++ * reciprocal;
			}
		}
		break;
	
	case AE_UNCH:
		DPRINTFZ(("Converting %d samples -- AE_UNCH\n", nSamples);)
		{	const unsigned char *pSample = (const unsigned char *)pBuffer;
			const float reciprocal = 1.0f / 128.0f;
			while (nSamples--) {
				*pOut++ = (float)*pSample++ * reciprocal - 1.0f;
			}
		}
		break;
	
	case AE_ULAW:
		DPRINTFZ(("Converting %d samples -- AE_ULAW\n", nSamples);)
		{	const unsigned char *pSample = (const unsigned char *)pBuffer;
			const float reciprocal = 1.0f / 32768.0f;
			while (nSamples--) {
				*pOut++ = (float)st_ulaw_to_linear(*pSample++) * reciprocal;
			}
		}
		break;
	
	case AE_ALAW:
		DPRINTFZ(("Converting %d samples -- AE_ALAW\n", nSamples);)
		{	const unsigned char *pSample = (const unsigned char *)pBuffer;
			const float reciprocal = 1.0f / 32768.0f;
			while (nSamples--) {
				*pOut++ = (float)st_alaw_to_linear(*pSample++) * reciprocal;
			}
		}
		break;
	
	case AE_SHORT:
		DPRINTFZ(("Converting %d samples -- AE_SHORT\n", nSamples);)
		{	const short *pSample = (const short *)pBuffer;
			const float reciprocal = 1.0f / 32768.0f;
			while (nSamples--) {
				*pOut++ = (float)*pSample++ * reciprocal;
			}
		}
		break;
	
	case AE_LONG:
		DPRINTFZ(("Converting %d samples -- AE_LONG\n", nSamples);)
		{	const long *pSample = (const long *)pBuffer;
			const float reciprocal = 1.0f / 32768.0f;
			while (nSamples--) {
				*pOut++ = (float)*pSample++ * reciprocal;
			}
		}
		break;
	
	case AE_FLOAT:
		DPRINTFZ(("Converting %d samples -- AE_FLOAT\n", nSamples);)
		{	const float *pSample = (const float *)pBuffer;
			const float reciprocal = 1.0f / 32768.0f;
			while (nSamples--) {
				*pOut++ = *pSample++ * reciprocal;
			}
		}
		break;
	}
}
