//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The transcript window starts and monitors a running Csound process.

#include <csignal>
#include <cstdio>
#include <cstring>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>

#include <Application.h>
#include <Font.h>
#include <KernelKit.h>
#include <Looper.h>
#include <Message.h>
#include <Path.h>
#include <Point.h>
#include <Rect.h>
#include <ScrollView.h>
#include <Window.h>

#include "CsApplication.h"
#include "CsAbstractJob.h"
#include "CsTranscriptView.h"
#include "CsTranscriptWindow.h"
#include "CsEnvJob.h"
#include "CsApplication.h"

// Message fields.
static const char kFieldSource[] = "source";

// Standard input given to children.
int CsoundTranscriptWindow::smBadStdin = -1;

// Initial dimensions of transcript window in characters.
const int kTextCols = 80;
const int kTextRows = 25;
const int kTabWidth = 8;

static BRect
initialViewFrame(float &tabWidth)
{
	const float colSpacing = be_fixed_font->StringWidth("M");
	tabWidth = kTabWidth * colSpacing;
	
	font_height fontHeight;
	be_fixed_font->GetHeight(&fontHeight);
	const float rowSpacing =
		fontHeight.ascent + fontHeight.descent + fontHeight.leading;

	BRect frame(B_ORIGIN, BPoint(kTextCols*colSpacing, kTextRows*rowSpacing));
	return frame;
}

CsoundTranscriptWindow::CsoundTranscriptWindow(
	BPoint origin, const char *pTitle, BLooper &target,
	const char *pExecutable, const CsoundAbstractJob &job)
:	inherited(BRect(origin, origin), pTitle, B_DOCUMENT_WINDOW, 0),
	mTarget(target),
	mChild(B_BAD_THREAD_ID),
	mChildFinished(create_sem(0, pTitle))
{
	// Get path of the executable. ### PRELIMINARY
	BPath path = CsoundApplication::GetDirectory();
	if (path.InitCheck() == B_NO_ERROR) {
		path.Append(pExecutable);
	} else {
		path = BPath(pExecutable);
	}

	// Get current directory, arguments and environment.
	BPath currentDir = job.CurrentDirectory();
	int32 nArgs;
	char **pArgv = job.CopyCommandArguments(path.Path(), nArgs);
	char **pEnv = CsoundApplication::EnvironmentJob().CopyEnvironment();

	// Set up a bad standard input that reads from a broken pipe.
	if (smBadStdin == -1) {
		// Make a file descriptor that is the read end of a broken pipe.
		int pd[2];
		if (pipe(pd) == 0) {
			// Close the write side.
			(void)close(pd[1]);
			// Save the read side.
			smBadStdin = pd[0];
		}
		if (smBadStdin != pd[0]) {
			// Oops - race condition.  No matter, we'll use the winner's FD.
			close(pd[0]);
		}
	}
	
	// Set up a pseudo-terminal to catch the child's output.
	int pd[2];
	if (!OpenPTY(pd[0], pd[1])) {
		// Oops.
		mTarget.PostMessage(kJobFailed);

	} else if (be_app->Lock()) {
		// In theory, *everybody* who uses the standard file descriptors must lock
		// the application to make this completely safe.

		// 1. Save the current file descriptors.
		int old_stdin = dup(0);
		int old_stdout = dup(1);
		int old_stderr = dup(2);
		
		// 2. Install the 'bad' standard input and the new standard output/error.
		dup2(smBadStdin, 0);
		dup2(pd[1], 1);
		dup2(pd[1], 2);
		close(pd[1]);

		// 3. Spawn the child process.
		chdir(currentDir.Path());
		mChild = load_image(nArgs,
								  const_cast<const char **>(pArgv),
								  const_cast<const char **>(pEnv));

		// 4. Restore original file descriptors.
		dup2(old_stdin, 0);
		close(old_stdin);		
		dup2(old_stdout, 1);
		close(old_stdout);		
		dup2(old_stderr, 2);
		close(old_stderr);		

		// Leave the critical section.
		be_app->Unlock();

		if (mChild > B_ERROR) {
			// Get a FILE pointer for the child's output.
			FILE *fp = fdopen(pd[0], "r");
			
			// Create a transcript view for the file.
			float tabWidth;
			CsoundTranscriptView *const pTranscript =
				new CsoundTranscriptView(initialViewFrame(tabWidth),
												 pTitle, fp, B_FOLLOW_ALL_SIDES, 0);
			pTranscript->MakeEditable(false);
			pTranscript->SetFontAndColor(be_fixed_font);
			pTranscript->SetTabWidth(tabWidth);
	
			// Create a scroll view around the transcript and make it fill
			// the window.
			BScrollView *const pScroller =
				new BScrollView("scroller", pTranscript, B_FOLLOW_ALL_SIDES, 0,
									 false, true, B_NO_BORDER);
			BRect frame = pScroller->Frame();
			ResizeTo(frame.Width()-1, frame.Height());
			AddChild(pScroller);
			
			// Show the window.
			Show();
			
			// Start up a thread which will notify us when the child terminates.
			thread_id monitor =
				spawn_thread(MonitorStub, pTitle, B_LOW_PRIORITY, this);
			if (monitor >= B_NO_ERROR) {
				// The thread function will cause the child process to start.
				resume_thread(monitor);
			}
			// ### what to do on failure?
				
		} else {
			// Child process could not be started.
			close(pd[0]);
			mTarget.PostMessage(kJobFailed);
		}
	}
	
	// Clean up.
	delete [] pArgv[0];
	delete [] pArgv;
	delete [] pEnv[0];
	delete [] pEnv;
}

CsoundTranscriptWindow::~CsoundTranscriptWindow(void)
{
	// Kill child if it is still running.  This shouldn't happen, but...
	KillChild();

	delete_sem(mChildFinished);
}

// Hook functions.
void
CsoundTranscriptWindow::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {
	
	// Force child process to terminate.
	case kAbortJob:
		KillChild();
		break;
	
	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

bool
CsoundTranscriptWindow::QuitRequested(void)
{
	KillChild();
	return true;
}

// Open a pseudo-terminal.
bool
CsoundTranscriptWindow::OpenPTY(int &pty, int &tty)
{
	if (DIR *pDir = opendir("/dev/tt")) {
		while (struct dirent *pEntry = readdir(pDir)) {
			if (strcmp(pEntry->d_name, ".") && strcmp(pEntry->d_name, "..")) {
				char buf[100];

				sprintf(buf, "/dev/pt/%s", pEntry->d_name);
				if ((pty = open(buf, O_RDWR)) < 0) {
					continue;
				}
				
				sprintf(buf, "/dev/tt/%s", pEntry->d_name);
				if ((tty = open(buf, O_RDWR)) < 0) {
					close(pty);
					continue;
				}
				
				closedir(pDir);
				return true;
			}
		}
		
		closedir(pDir);
	}
	
	return false;
}

// Monitor the child process.
int32
CsoundTranscriptWindow::MonitorStub(void *pData)
{
	return ((CsoundTranscriptWindow *)pData)->Monitor();
}

int32
CsoundTranscriptWindow::Monitor(void)
{
	// Wait for the child to finish.
	status_t ret, status;
	while ((ret = wait_for_thread(mChild, &status)) < B_NO_ERROR
			 && ret == B_INTERRUPTED)
		;

	// OK, clean up and signal the target that we're done.
	mChild = B_BAD_THREAD_ID;
	mTarget.PostMessage(kJobFinished);
	release_sem(mChildFinished);

	return 0;
}

// Kill the child process.
void
CsoundTranscriptWindow::KillChild(void)
{
	if (mChild > B_ERROR) {
		send_signal(mChild, SIGKILL);
		acquire_sem(mChildFinished);
	}
}
