//		Copyright © 1997–2001  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Application class for Csound.

#include <Alert.h>
#include <Application.h>
#include <Archivable.h>
#include <Entry.h>
#include <File.h>
#include <FilePanel.h>
#include <FindDirectory.h>
#include <List.h>
#include <Menu.h>
#include <MenuItem.h>
#include <Path.h>
#include <Roster.h>
#include <Window.h>

#include "CsApplication.h"
#include "CsAbstractJob.h"
#include "CsSynthJob.h"	
#include "CsHFJob.h"	
#include "CsLPJob.h"	
#include "CsPVJob.h"	
#include "CsCVJob.h"	
#include "CsEnvJob.h"	
#include "CsAbstractWindow.h"	
#include "CsAbstractMainWindow.h"	
#include "CsEnvMainWindow.h"	
#include "CsSvInterface.h"	

// Application signature.
static const char kSignature[] = "application/x-mit-csound";

// Label strings.
static const char kWindowEnvironmentLabel[] = "Environment";

// Message fields.
static const char kFieldRefs[] = "refs";
const char CsoundApplication::kFieldWindow[] = "CsWindow";

// File names.
static const char kDefaultEnvironmentFile[] = "CsoundEnvironment";

BWindow *CsoundApplication::smpEnvironmentWindow = (BWindow *)0;

// BApplication requirements.
CsoundApplication::CsoundApplication(void)
:	inherited(kSignature),
	mpFilePanel((BFilePanel *)0),
	mFileSelectionPending(false),
	mWindowOpened(false),
	mPrototypeJobs(),
	mMainWindows()
{
	for (int i = 0; i < kNumShortcuts; ++i) {
		mpShortcutWindows[i] = 0;
	}
	
	// Register prototype jobs.  Every new concrete job class must be added here.
	mPrototypeJobs.AddItem(new CsoundSynthesisJob(false));
	mPrototypeJobs.AddItem(new CsoundHFAnalysisJob(false));
	mPrototypeJobs.AddItem(new CsoundLPAnalysisJob(false));
	mPrototypeJobs.AddItem(new CsoundPVAnalysisJob(false));
	mPrototypeJobs.AddItem(new CsoundCVAnalysisJob(false));
	//mPrototypeJobs.AddItem(new CsoundSoundinfoJob(false));
}

CsoundApplication::~CsoundApplication(void)
{
	delete mpFilePanel;
}

// Hook functions.
void
CsoundApplication::AboutRequested(void)
{
	(new BAlert("About",
					"Csound shell for BeOS\n"
					"Copyright " B_UTF8_COPYRIGHT " 1997–2000  Jens Kilian",
					"OK"))->Go();
}

void
CsoundApplication::ArgvReceived(int32 argc, char **ppArgv)
{
	// Loop through all command-line arguments and treat them as file names.
	while (--argc > 0) {
		OpenFile(BEntry(*++ppArgv));
	}
}

void
CsoundApplication::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {

	// Update the window list.	
	case kMainWindowCreated:
		{	void *pPointer;
			if (pMsg->FindPointer(kFieldWindow, &pPointer) == B_NO_ERROR) {
				BWindow *pWindow = static_cast<BWindow *>(pPointer);
				if (pWindow && pWindow != smpEnvironmentWindow) {
					mMainWindows.AddItem(pWindow);
					SetShortcut(0, pWindow);
				}
			}
		}
		break;
	
	case kMainWindowDestroyed:
		{	void *pPointer;
			if (pMsg->FindPointer(kFieldWindow, &pPointer) == B_NO_ERROR) {
				BWindow *pWindow = static_cast<BWindow *>(pPointer);
				if (pWindow && pWindow != smpEnvironmentWindow) {
					mMainWindows.RemoveItem(pWindow);
					SetShortcut(pWindow, 0);
					
					// Quit after last main window was closed.
					if (mMainWindows.CountItems() < 1) {
						PostMessage(B_QUIT_REQUESTED);
					}
				}
			}
		}
		break;
	
	// Open a file.  Actually, just show the file panel :-)
	case kOpenFile:
		ShowFilePanel();
		break;
	
	// File panel closed.
	case B_CANCEL:
		mFileSelectionPending = false;
		break;
		
	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

bool
CsoundApplication::TryClosingWindow(void *pWin, void *pFlag)
{
	BWindow *pWindow = (BWindow *)pWin;
	if (pWindow->Lock()) {
		if (pWindow->QuitRequested()) {
			pWindow->Quit();

		} else {
			pWindow->Unlock();
			// Can't quit the application.
			*(bool *)pFlag = false;
		}
	}
	
	return false;
}

bool
CsoundApplication::QuitRequested(void)
{
	// Don't allow quitting while a file selection is pending.
	if (mFileSelectionPending) {
		ShowFilePanel();	// it might have been minimized...
		return false;
	}
	
	// Try to Quit() all open main windows.
	bool allowed = true;
	mMainWindows.DoForEach(TryClosingWindow, &allowed);
	
	if (allowed && smpEnvironmentWindow && smpEnvironmentWindow->Lock()) {
		// Try to save the environment settings.
		BPath path;
		if (find_directory(B_USER_SETTINGS_DIRECTORY, &path, true) == B_OK
			 && path.Append(kDefaultEnvironmentFile) == B_OK) {
			BEntry entry(path.Path());
			entry_ref ref;
			if (entry.GetRef(&ref) == B_NO_ERROR) {
				CsoundEnvironmentJob &job =
					const_cast<CsoundEnvironmentJob &>(EnvironmentJob());
				job.Save(&ref);
			}
		}

		// OK, now also dispose of the environment window.
		smpEnvironmentWindow->Quit();
		smpEnvironmentWindow = (BWindow *)0;
	}

	return allowed;
}

void
CsoundApplication::ReadyToRun(void)
{
	// Try to load a default environment (if no environment job was loaded yet).
	if (!smpEnvironmentWindow) {
		BPath path;
		if (find_directory(B_USER_SETTINGS_DIRECTORY, &path) == B_OK
			 && path.Append(kDefaultEnvironmentFile) == B_OK) {
			OpenFile(BEntry(path.Path()));
		}
	}
	if (!smpEnvironmentWindow) {
		BPath path;
		if (find_directory(B_COMMON_SETTINGS_DIRECTORY, &path) == B_OK
			 && path.Append(kDefaultEnvironmentFile) == B_OK) {
			OpenFile(BEntry(path.Path()));
		}
	}
	
	// Create the environment window if no default environment was found.
	if (!smpEnvironmentWindow) {
		SetEnvironment(CsoundEnvironmentJob(false));
	}
	
	// Open an initial window if no files were loaded.
	if (!mWindowOpened) {
		CsoundSynthesisJob().CreateWindow(BPoint(100, 100))->Show();
	}
	
	// Try to start the Csound Server if it isn't already running.
	status_t status = be_roster->Launch(kServerSignature);
	if (status != B_OK && status != B_ALREADY_RUNNING) {
		(new BAlert("Server Failure",
						"Failed to start Csound Server.\n"
						"Graphics displays will not be available.",
						"Sorry"))->Go();
	}
}

void
CsoundApplication::RefsReceived(BMessage *pMsg)
{
	// Loop through all the references.
	type_code type;
	int32 count;
	if (pMsg->GetInfo(kFieldRefs, &type, &count) == B_NO_ERROR) {
		for (int32 i = --count; i >= 0; i--) {
			entry_ref ref;
			if (pMsg->FindRef(kFieldRefs, i, &ref) == B_NO_ERROR) {
				OpenFile(BEntry(&ref));
			}
		}
	}
}

// Return the application signature.
const char *
CsoundApplication::GetSignature(void)
{
	return kSignature;
}

// Return the application directory.
BPath
CsoundApplication::GetDirectory(void)
{
	BPath path;
	
	app_info info;
	if (be_app->GetAppInfo(&info) == B_NO_ERROR) {
		BEntry appEntry(&(info.ref));
		if (appEntry.InitCheck() == B_NO_ERROR
			 && appEntry.GetParent(&appEntry) == B_NO_ERROR) {
			appEntry.GetPath(&path);
		}
	}
	
	return path;
}

// Show our file panel.
void
CsoundApplication::ShowFilePanel(void)
{
	if (!mpFilePanel) {
		// Create the file panel.
		mpFilePanel = new BFilePanel;
	}

	mpFilePanel->Show();
	mFileSelectionPending = true;
}

// Try to open a file and retrieve a saved job.
void
CsoundApplication::OpenFile(const BEntry &entry)
{
	if (entry.IsFile()) {
		BFile theFile(&entry, B_READ_ONLY);
		BMessage msg;
		if (msg.Unflatten(&theFile) == B_NO_ERROR) {
			BArchivable *pArchived = instantiate_object(&msg);
			CsoundAbstractJob *const pJob =
				dynamic_cast<CsoundAbstractJob *>(pArchived);
			if (pJob) {
				CsoundEnvironmentJob *const pEnvironment =
					dynamic_cast<CsoundEnvironmentJob *>(pJob);
				if (pEnvironment) {
					// Special case - set environment only.
					SetEnvironment(*pEnvironment);
	
				} else {
					// Open a window for the loaded job.
					entry_ref ref;
					if (entry.GetRef(&ref) == B_NO_ERROR) {
						pJob->SetEntry(ref);
						pJob->CreateWindow(BPoint(100, 100))->Show();
						mWindowOpened = true;
					}
				}
			}
			delete pArchived;
		}
	}
}

// Return a job describing environment variables.
const CsoundEnvironmentJob &
CsoundApplication::EnvironmentJob(void)
{
	const CsoundEnvironmentMainWindow &window =
		dynamic_cast<CsoundEnvironmentMainWindow &>(*smpEnvironmentWindow);
	return dynamic_cast<const CsoundEnvironmentJob &>(window.CurrentJob());
}

// Set up or modify the current environment.
void
CsoundApplication::SetEnvironment(const CsoundEnvironmentJob &job) const
{
	if (!smpEnvironmentWindow) {
		// Create the environment window.
		smpEnvironmentWindow = job.CreateWindow(BPoint(200, 200));
		smpEnvironmentWindow->Run();
	} else if (smpEnvironmentWindow->Lock()) {
		// Modify environment.
		CsoundEnvironmentMainWindow &window =
			dynamic_cast<CsoundEnvironmentMainWindow &>(*smpEnvironmentWindow);
		window.SetCurrentJob(job);
		window.Unlock();
	}
}

// Construct a menu for switching between open windows.
int
CsoundApplication::ShortcutForWindow(BWindow *pWindow)
{
	for (int i = 0; i < kNumShortcuts; ++i) {
		if (mpShortcutWindows[i] == pWindow) {
			return i;
		}
	}

	return -1;
}

void
CsoundApplication::SetShortcut(BWindow *pOld, BWindow *pNew)
{
	const int shortcut = ShortcutForWindow(pOld);
	if (shortcut >= 0) {
		mpShortcutWindows[shortcut] = pNew;
	}
}

bool
CsoundApplication::CreateWindowMenuItem(void *pWin, void *pMenu)
{
	CsoundApplication *pApp = (CsoundApplication *)be_app;
	CsoundAbstractMainWindow *pWindow = (CsoundAbstractMainWindow *)pWin;

	const int shortcut = pApp->ShortcutForWindow(pWindow);
	BMenuItem *pMenuItem =
		new BMenuItem(pWindow->CurrentJob().JobFilename(),
						  new BMessage(CsoundAbstractMainWindow::kShowWindow),
						  (shortcut < 0) ? 0 : '0' + shortcut);
	pMenuItem->SetTarget(pWindow);
	((BMenu *)pMenu)->AddItem(pMenuItem);

	return false;
}

void
CsoundApplication::ConstructWindowMenu(BMenu &menu)
{
	// Remove current menu entries.
	int32 i = menu.CountItems();
	while (i > 0) {
		delete menu.RemoveItem(--i);
	}

	if (Lock()) {
		// Add the special entry for the environment window.
		if (smpEnvironmentWindow) {
			BMenuItem *pMenuItem =
				new BMenuItem(kWindowEnvironmentLabel,
								  new BMessage(CsoundAbstractMainWindow::kShowWindow),
								  'E');
			pMenuItem->SetTarget(smpEnvironmentWindow);
			menu.AddItem(pMenuItem);
			menu.AddSeparatorItem();
		}
	
		// Add entries for all open main windows.
		mMainWindows.DoForEach(CreateWindowMenuItem, &menu);

		Unlock();
	}	
}
